/*
 * Generation of declaration elements such as parameters and members.
 *
 * Copyright (C) 2019 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include "common.h"
#include "config.h"
#include "declaration.h"
#include "templates.h"



/* Generate each operation signature for use in defining an interface. */

void write_interface_signature(struct signature *sig, enum signature_role role,
                               enum component_role component, FILE *fp,
                               struct interface *iface)
{
  int cpp = (conf.language == CPP_LANGUAGE);
  int have_init_param = (conf.language == C_LANGUAGE);
  int declaration = (role & DECLARATION_ROLE);
  int member = (role & MEMBER_ROLE);
  int server = (component == SERVER_ROLE);
  int completion = have_attribute(sig->attributes, "completion");
  enum specifier direction = (server && completion) ? IN_PARAMETER : ANY_PARAMETER;
  enum function_role function = (server && completion) ? COMPLETION_ROLE : GENERAL_FUNCTION_ROLE;

  /* Indent and provide a qualifier for C++ class declarations. */

  char *qualifier = (cpp && declaration) ? "  virtual " : member ? "  " : "";

  /* Structure the function name appropriately. */

  char *opname = get_signature_operation_name(iface, sig, component, role,
                                              conf.language);

  /* Introduce an initial parameter for the component state. */

  if (!cpp && member)
    fprintf(fp, interface_struct_member_function_signature_prologue, opname);
  else
    fprintf(fp, interface_function_signature_prologue, qualifier, opname);

  if (have_init_param)
    write_initial_parameter(iface->name, fp);

  /* Emit the parameters. */

  write_parameters(sig->parameters, fp, SIGNATURE_ROLE, function, direction,
                   have_init_param);

  fputs(")", fp);

  /* Add a body for the function, this being used to define unimplemented
     methods in C++. The client has method definitions and therefore normal
     virtual methods. */

  if (cpp && declaration && server)
    write_interface_unimplemented_method(sig, direction, fp);
  else
    fputs(get_signature_terminator(role), fp);

  free(opname);
}

/* Write an unimplemented member function body for a C++ interface. */

void write_interface_unimplemented_method(struct signature *sig,
                                          enum specifier direction, FILE *fp)
{
  struct parameter *param;

  fputs(interface_unimplemented_method_prologue_cpp, fp);

  for (param = sig->parameters; param != NULL; param = param->tail)
  {
    if (param->specifier & direction)
      fprintf(fp, interface_unused_parameter, get_parameter_name(param->identifiers));
  }

  fputs(interface_unimplemented_method_epilogue_cpp, fp);
}

/* Write the initial parameter in an interface signature. */

void write_initial_parameter(const char *name, FILE *fp)
{
  if (conf.language == CPP_LANGUAGE)
    fprintf(fp, "%s *_self", name);
  else
    fprintf(fp, "ref_%s _self", name);
}

/* Generate signatures. */

void write_signatures(struct signature *sig, enum signature_role role, FILE *fp,
                      struct interface *iface,
                      void (*write_signature)(struct signature *,
                                              enum signature_role role,
                                              FILE *, struct interface *))
{
  if (sig == NULL)
    return;

  write_signature(sig, role, fp, iface);
  write_signatures(sig->tail, role, fp, iface, write_signature);
}



/* Generate parameters in each function signature. */

void write_parameters(struct parameter *param, FILE *fp, enum parameter_role role,
                      enum function_role function, enum specifier direction,
                      int continuing)
{
  for (; param != NULL; param = param->tail)
  {
    if (param->specifier & direction)
    {
      if (continuing)
        fputs(", ", fp);
      else
        continuing = 1;

      write_parameter(param, fp, role, function);
    }
  }
}

/* Generate a parameter in a signature or invocation. */

void write_parameter(struct parameter *param, FILE *fp,
                     enum parameter_role role, enum function_role function)
{
  struct identifier *ident = param->identifiers;
  char *type;
  int invocation = (role == INVOCATION_ROLE);
  int first = 1;

  while (ident != NULL)
  {
    /* Put spaces between each type identifier. */

    if (!invocation)
    {
      if (first)
        first = 0;
      else
        fputs(" ", fp);
    }

    /* Emit type identifiers for signatures. */

    if (ident->tail != NULL)
    {
      if (!invocation)
        fputs(ident->identifier, fp);
    }

    /* Emit the parameter name. */

    else
    {
      /* Items must have their type identifiers restored. */

      if ((param->cls & ITEM_CLASS) && !invocation)
      {
        type = param->cls == FPAGE_ITEM ? L4_FPAGE_TYPE : L4_CAP_TYPE;
        fputs(type, fp);
        fputs(" ", fp);
      }

      /* Emit the decoration and parameter name. */

      fprintf(fp, "%s%s", get_parameter_decoration(param, role, function),
                          ident->identifier);
    }

    ident = ident->tail;
  }
}

/* Write structure members for word data structures or variable declarations for
   parameters. */

void write_declarations(struct parameter *param, enum specifier direction,
                        enum parameter_class cls, FILE *fp)
{
  for (; param != NULL; param = param->tail)
  {
    if ((param->specifier & direction) && (param->cls & cls))
    {
      fputs(INDENT_MEMBER, fp);
      write_parameter(param, fp, STRUCTURE_ROLE, GENERAL_FUNCTION_ROLE);
      fputs(COMPLETE_MEMBER, fp);
    }
  }
}
